package rtr;

import addr.addrIP;
import cfg.cfgAll;
import ip.ipFwdIface;
import ip.ipFwdTab;
import java.util.Comparator;
import java.util.List;
import pipe.pipeLine;
import pipe.pipeSide;
import prt.prtAccept;
import tab.tabPlcmapN;
import tab.tabPrfxlstN;
import tab.tabRoute;
import tab.tabRouteEntry;
import tab.tabRtrmapN;
import tab.tabRtrplc;
import util.bits;
import util.debugger;
import util.logger;

/**
 * bgp4 neighbor
 *
 * @author matecsaba
 */
public class rtrBgpNeigh extends rtrBgpParam implements Comparator<rtrBgpNeigh>, Runnable {

    /**
     * address of peer
     */
    public addrIP peerAddr;

    /**
     * local interface
     */
    public ipFwdIface localIfc = null;

    /**
     * local address
     */
    public addrIP localAddr = new addrIP();

    /**
     * accepted unicast prefixes
     */
    public tabRoute<addrIP> accUni = new tabRoute<addrIP>("rx");

    /**
     * accepted multicast prefixes
     */
    public tabRoute<addrIP> accMlt = new tabRoute<addrIP>("rx");

    /**
     * accepted flowspec prefixes
     */
    public tabRoute<addrIP> accFlw = new tabRoute<addrIP>("rx");

    /**
     * accepted vpnuni prefixes
     */
    public tabRoute<addrIP> accVpnU = new tabRoute<addrIP>("rx");

    /**
     * accepted vpnmulti prefixes
     */
    public tabRoute<addrIP> accVpnM = new tabRoute<addrIP>("rx");

    /**
     * accepted vpnflow prefixes
     */
    public tabRoute<addrIP> accVpnF = new tabRoute<addrIP>("rx");

    /**
     * accepted vpls prefixes
     */
    public tabRoute<addrIP> accVpls = new tabRoute<addrIP>("rx");

    /**
     * accepted evpn prefixes
     */
    public tabRoute<addrIP> accEvpn = new tabRoute<addrIP>("rx");

    /**
     * accepted mdt prefixes
     */
    public tabRoute<addrIP> accMdt = new tabRoute<addrIP>("rx");

    /**
     * willing unicast prefixes
     */
    public tabRoute<addrIP> wilUni = new tabRoute<addrIP>("tx");

    /**
     * willing multicast prefixes
     */
    public tabRoute<addrIP> wilMlt = new tabRoute<addrIP>("tx");

    /**
     * willing flowspec prefixes
     */
    public tabRoute<addrIP> wilFlw = new tabRoute<addrIP>("tx");

    /**
     * willing vpnuni prefixes
     */
    public tabRoute<addrIP> wilVpnU = new tabRoute<addrIP>("tx");

    /**
     * willing vpnmulti prefixes
     */
    public tabRoute<addrIP> wilVpnM = new tabRoute<addrIP>("tx");

    /**
     * willing vpnflow prefixes
     */
    public tabRoute<addrIP> wilVpnF = new tabRoute<addrIP>("tx");

    /**
     * willing vpls prefixes
     */
    public tabRoute<addrIP> wilVpls = new tabRoute<addrIP>("tx");

    /**
     * willing evpn prefixes
     */
    public tabRoute<addrIP> wilEvpn = new tabRoute<addrIP>("tx");

    /**
     * willing mdt prefixes
     */
    public tabRoute<addrIP> wilMdt = new tabRoute<addrIP>("tx");

    /**
     * changed unicast prefixes
     */
    public tabRoute<addrIP> chgUni = new tabRoute<addrIP>("chg");

    /**
     * changed multicast prefixes
     */
    public tabRoute<addrIP> chgMlt = new tabRoute<addrIP>("chg");

    /**
     * changed flowspec prefixes
     */
    public tabRoute<addrIP> chgFlw = new tabRoute<addrIP>("chg");

    /**
     * changed vpnuni prefixes
     */
    public tabRoute<addrIP> chgVpnU = new tabRoute<addrIP>("chg");

    /**
     * changed vpnmulti prefixes
     */
    public tabRoute<addrIP> chgVpnM = new tabRoute<addrIP>("chg");

    /**
     * changed vpnflow prefixes
     */
    public tabRoute<addrIP> chgVpnF = new tabRoute<addrIP>("chg");

    /**
     * changed vpls prefixes
     */
    public tabRoute<addrIP> chgVpls = new tabRoute<addrIP>("chg");

    /**
     * changed evpn prefixes
     */
    public tabRoute<addrIP> chgEvpn = new tabRoute<addrIP>("chg");

    /**
     * changed mdt prefixes
     */
    public tabRoute<addrIP> chgMdt = new tabRoute<addrIP>("chg");

    /**
     * update group number
     */
    public int groupMember = -1;

    /**
     * neighbor reachable
     */
    public boolean reachable = false;

    /**
     * old reachable state
     */
    public boolean reachOld;

    /**
     * neighbor reachable change time
     */
    public long reachTim;

    /**
     * neighbor reachable change number
     */
    public long reachNum;

    /**
     * speaker
     */
    public rtrBgpSpeak conn;

    /**
     * type of peer
     */
    public int peerType;

    /**
     * full compute last
     */
    public long fullLast;

    /**
     * incremental compute last
     */
    public long incrLast;

    /**
     * full compute count
     */
    public int fullCount;

    /**
     * incremental compute count
     */
    public int incrCount;

    /**
     * full compute time
     */
    public int fullTime;

    /**
     * incremental compute time
     */
    public int incrTime;

    private boolean need2run;

    /**
     * create neighbor
     *
     * @param parent bgp process
     */
    public rtrBgpNeigh(rtrBgp parent) {
        super(parent, false);
        peerAddr = new addrIP();
        conn = new rtrBgpSpeak(lower, this, null);
    }

    public int compare(rtrBgpNeigh o1, rtrBgpNeigh o2) {
        return o1.peerAddr.compare(o1.peerAddr, o2.peerAddr);
    }

    public String toString() {
        return "" + peerAddr;
    }

    public void flapBgpConn() {
        conn.closeNow();
    }

    public void doTempCfg(String cmd, boolean negated) {
    }

    public void getConfig(List<String> l, String beg, boolean filter) {
        l.addAll(getParamCfg(beg, "neighbor " + peerAddr + " ", filter));
    }

    /**
     * get status of peer
     *
     * @param l list to append
     */
    public void getStatus(List<String> l) {
        l.add("peer = " + peerAddr);
        l.add("reachable state = " + reachable);
        l.add("reachable changed = " + bits.timePast(reachTim));
        l.add("reachable changes = " + reachNum);
        l.add("update group = " + groupMember);
        l.add("type = " + rtrBgpUtil.peerType2string(peerType));
        l.add("safi = " + rtrBgpParam.mask2string(conn.peerAfis));
        l.add("local = " + localAddr);
        l.add("router id = " + conn.peerRouterID);
        l.add("uptime = " + bits.timePast(conn.upTime));
        l.add("hold time = " + bits.timeDump(conn.peerHold / 1000));
        l.add("keepalive time = " + bits.timeDump(conn.peerKeep / 1000));
        l.add("32bit as = " + conn.peer32bitAS);
        l.add("refresh = " + conn.peerRefresh);
        l.add("graceful = " + rtrBgpParam.mask2string(conn.peerGrace));
        l.add("addpath rx = " + rtrBgpParam.mask2string(conn.addpathRx));
        l.add("addpath tx = " + rtrBgpParam.mask2string(conn.addpathTx));
        l.add("unicast advertised = " + conn.advUni.size() + " of " + wilUni.size() + ", list=" + chgUni.size());
        l.add("multicast advertised = " + conn.advMlt.size() + " of " + wilMlt.size() + ", list=" + chgMlt.size());
        l.add("flowspec advertised = " + conn.advFlw.size() + " of " + wilFlw.size() + ", list=" + chgFlw.size());
        l.add("vpnuni advertised = " + conn.advVpnU.size() + " of " + wilVpnU.size() + ", list=" + chgVpnU.size());
        l.add("vpnmlt advertised = " + conn.advVpnM.size() + " of " + wilVpnM.size() + ", list=" + chgVpnM.size());
        l.add("vpnflw advertised = " + conn.advVpnF.size() + " of " + wilVpnF.size() + ", list=" + chgVpnF.size());
        l.add("vpls advertised = " + conn.advVpls.size() + " of " + wilVpls.size() + ", list=" + chgVpls.size());
        l.add("evpn advertised = " + conn.advEvpn.size() + " of " + wilEvpn.size() + ", list=" + chgEvpn.size());
        l.add("mdt advertised = " + conn.advMdt.size() + " of " + wilMdt.size() + ", list=" + chgMdt.size());
        l.add("unicast accepted = " + accUni.size() + " of " + conn.lrnUni.size());
        l.add("multicast accepted = " + accMlt.size() + " of " + conn.lrnMlt.size());
        l.add("flowspec accepted = " + accFlw.size() + " of " + conn.lrnFlw.size());
        l.add("vpnuni accepted = " + accVpnU.size() + " of " + conn.lrnVpnU.size());
        l.add("vpnmlt accepted = " + accVpnM.size() + " of " + conn.lrnVpnM.size());
        l.add("vpnflw accepted = " + accVpnF.size() + " of " + conn.lrnVpnF.size());
        l.add("vpls accepted = " + accVpls.size() + " of " + conn.lrnVpls.size());
        l.add("evpn accepted = " + accEvpn.size() + " of " + conn.lrnEvpn.size());
        l.add("mdt accepted = " + accMdt.size() + " of " + conn.lrnMdt.size());
        l.add("version = " + conn.adversion + " of " + lower.compRound + ", need=" + conn.needFull + ", init=" + conn.initFull);
        l.add("full = " + fullCount + ", " + bits.time2str(cfgAll.timeZoneName, fullLast + cfgAll.timeServerOffset, 3) + ", " + bits.timePast(fullLast) + " ago, " + fullTime + " ms");
        l.add("incr = " + incrCount + ", " + bits.time2str(cfgAll.timeZoneName, incrLast + cfgAll.timeServerOffset, 3) + ", " + bits.timePast(incrLast) + " ago, " + incrTime + " ms");
        l.add("connection = " + conn.cntr.getShStat());
        l.add("buffer = " + pipeSide.getStatus(conn.pipe));
    }

    /**
     * update peer structures
     */
    public void updatePeer() {
        if (localAs != remoteAs) {
            peerType = rtrBgpUtil.peerExtrn;
            if (remoteConfed) {
                peerType = rtrBgpUtil.peerCnfed;
            }
            if (serverClnt) {
                peerType = rtrBgpUtil.peerServr;
            }
        } else {
            peerType = rtrBgpUtil.peerIntrn;
            if (reflectClnt) {
                peerType = rtrBgpUtil.peerRflct;
            }
        }
    }

    /**
     * start this neighbor process
     */
    public void startNow() {
        if (need2run) {
            return;
        }
        need2run = true;
        new Thread(this).start();
    }

    /**
     * stop this neighbor process
     */
    public void stopNow() {
        need2run = false;
        shutdown = true;
        conn.closeNow();
    }

    public void run() {
        try {
            doWork();
        } catch (Exception e) {
            logger.traceback(e);
        }
        need2run = false;
        conn.closeNow();
    }

    private void doWork() {
        int lastKeep = 0;
        for (;;) {
            bits.sleep(100);
            lastKeep++;
            if (lastKeep > (conn.peerKeep / 100)) {
                conn.sendKeepAlive();
                lastKeep = 0;
            }
            if (!need2run) {
                return;
            }
            if (shutdown) {
                continue;
            }
            if (conn.txFree() >= 0) {
                doAdvert();
                continue;
            }
            lastKeep = 0;
            switch (socketMode) {
                case 1: // active
                    openConn(0);
                    break;
                case 2: // passive
                    openConn(60);
                    break;
                case 3: // both
                    if (!openConn(bits.random(10, 30))) {
                        break;
                    }
                    openConn(0);
                    break;
                case 4: // dynamic
                    lower.lstnNei.del(this);
                    stopNow();
                    return;
                case 5: // deleted dynamic
                    return;
            }
        }
    }

    private boolean openConn(int tim) {
        ipFwdIface ifc;
        if (srcIface == null) {
            ifc = ipFwdTab.findSendingIface(lower.fwdCore, peerAddr);
        } else {
            ifc = srcIface.getFwdIfc(peerAddr);
        }
        if (ifc == null) {
            return true;
        }
        if (ifc.addr == null) {
            return true;
        }
        pipeSide pipe;
        if (tim < 1) {
            pipe = lower.tcpCore.streamConnect(new pipeLine(bufferSize, false), ifc, 0, peerAddr, rtrBgp.port, "bgp", passwd, ttlSecurity);
        } else {
            prtAccept ac = new prtAccept(lower.tcpCore, new pipeLine(bufferSize, false), ifc, rtrBgp.port, peerAddr, 0, 0, "bgp", passwd, ttlSecurity);
            ac.wait4conn(tim * 1000);
            pipe = ac.getConn(true);
        }
        if (pipe == null) {
            return true;
        }
        localIfc = ifc;
        localAddr = ifc.addr.copyBytes();
        conn.closeNow();
        conn = new rtrBgpSpeak(lower, this, pipe);
        return false;
    }

    private boolean advertFullTable(int safi, int mask, tabRoute<addrIP> will, tabRoute<addrIP> done) {
        if ((conn.peerAfis & mask) == 0) {
            return false;
        }
        if (conn.needFull < 2) {
            will = new tabRoute<addrIP>(will);
        }
        for (int i = 0; i < will.size(); i++) {
            tabRouteEntry<addrIP> ntry = will.get(i);
            if (ntry == null) {
                continue;
            }
            if (!ntry.differs(done.find(ntry))) {
                continue;
            }
            done.add(3, ntry, false, false);
            conn.sendUpdate(safi, ntry, true);
            if (conn.txFree() < 1024) {
                return true;
            }
        }
        for (int i = done.size() - 1; i >= 0; i--) {
            tabRouteEntry<addrIP> ntry = done.get(i);
            if (ntry == null) {
                continue;
            }
            if (will.find(ntry) != null) {
                continue;
            }
            done.del(ntry);
            conn.sendUpdate(safi, ntry, false);
            if (conn.txFree() < 1024) {
                return true;
            }
        }
        return false;
    }

    private boolean advertFull() {
        long tim = bits.getTime();
        if (advertFullTable(lower.afiUni, rtrBgpParam.mskUni, wilUni, conn.advUni)) {
            return true;
        }
        if (advertFullTable(lower.afiLab, rtrBgpParam.mskLab, wilUni, conn.advUni)) {
            return true;
        }
        if (advertFullTable(lower.afiMlt, rtrBgpParam.mskMlt, wilMlt, conn.advMlt)) {
            return true;
        }
        if (advertFullTable(lower.afiFlw, rtrBgpParam.mskFlw, wilFlw, conn.advFlw)) {
            return true;
        }
        if (advertFullTable(lower.afiVpnU, rtrBgpParam.mskVpnU, wilVpnU, conn.advVpnU)) {
            return true;
        }
        if (advertFullTable(lower.afiVpnM, rtrBgpParam.mskVpnM, wilVpnM, conn.advVpnM)) {
            return true;
        }
        if (advertFullTable(lower.afiVpnF, rtrBgpParam.mskVpnF, wilVpnF, conn.advVpnF)) {
            return true;
        }
        if (advertFullTable(lower.afiVpls, rtrBgpParam.mskVpls, wilVpls, conn.advVpls)) {
            return true;
        }
        if (advertFullTable(lower.afiEvpn, rtrBgpParam.mskEvpn, wilEvpn, conn.advEvpn)) {
            return true;
        }
        if (advertFullTable(lower.afiMdt, rtrBgpParam.mskMdt, wilMdt, conn.advMdt)) {
            return true;
        }
        if (conn.needFull == 1) {
            conn.needFull--;
        } else {
            conn.needFull = 1;
        }
        if (conn.initFull) {
            conn.needFull = 1;
        }
        conn.initFull = false;
        fullLast = bits.getTime();
        fullTime = (int) (fullLast - tim);
        fullCount++;
        return false;
    }

    private boolean advertIncrTable(int safi, int mask, tabRoute<addrIP> will, tabRoute<addrIP> chg, tabRoute<addrIP> done) {
        if ((conn.peerAfis & mask) == 0) {
            return false;
        }
        chg = new tabRoute<addrIP>(chg);
        for (int i = 0; i < chg.size(); i++) {
            tabRouteEntry<addrIP> cur = chg.get(i);
            if (cur == null) {
                continue;
            }
            tabRouteEntry<addrIP> wil = will.find(cur);
            tabRouteEntry<addrIP> don = done.find(cur);
            if (wil == null) {
                if (don == null) {
                    continue;
                }
                done.del(don);
                conn.sendUpdate(safi, don, false);
            } else {
                if (!wil.differs(don)) {
                    continue;
                }
                done.add(3, wil, false, false);
                conn.sendUpdate(safi, wil, true);
            }
            if (conn.txFree() < 1024) {
                return true;
            }
        }
        return false;
    }

    private boolean advertIncr() {
        long tim = bits.getTime();
        if (advertIncrTable(lower.afiUni, rtrBgpParam.mskUni, wilUni, chgUni, conn.advUni)) {
            return true;
        }
        if (advertIncrTable(lower.afiLab, rtrBgpParam.mskLab, wilUni, chgUni, conn.advUni)) {
            return true;
        }
        if (advertIncrTable(lower.afiMlt, rtrBgpParam.mskMlt, wilMlt, chgMlt, conn.advMlt)) {
            return true;
        }
        if (advertIncrTable(lower.afiFlw, rtrBgpParam.mskFlw, wilFlw, chgFlw, conn.advFlw)) {
            return true;
        }
        if (advertIncrTable(lower.afiVpnU, rtrBgpParam.mskVpnU, wilVpnU, chgVpnU, conn.advVpnU)) {
            return true;
        }
        if (advertIncrTable(lower.afiVpnM, rtrBgpParam.mskVpnM, wilVpnM, chgVpnM, conn.advVpnM)) {
            return true;
        }
        if (advertIncrTable(lower.afiVpnF, rtrBgpParam.mskVpnF, wilVpnF, chgVpnF, conn.advVpnF)) {
            return true;
        }
        if (advertIncrTable(lower.afiVpls, rtrBgpParam.mskVpls, wilVpls, chgVpls, conn.advVpls)) {
            return true;
        }
        if (advertIncrTable(lower.afiEvpn, rtrBgpParam.mskEvpn, wilEvpn, chgEvpn, conn.advEvpn)) {
            return true;
        }
        if (advertIncrTable(lower.afiMdt, rtrBgpParam.mskMdt, wilMdt, chgMdt, conn.advMdt)) {
            return true;
        }
        incrLast = bits.getTime();
        incrTime = (int) (incrLast - tim);
        incrCount++;
        return false;
    }

    private void doAdvert() {
        if ((bits.getTime() - conn.lastRx) > holdTimer) {
            conn.sendNotify(4, 0);
            conn.closeNow();
            return;
        }
        int doing = lower.compRound;
        if (doing == conn.adversion) {
            return;
        }
        if (conn.txFree() < (bufferSize / 2)) {
            return;
        }
        boolean b;
        if (conn.needFull != 0) {
            b = advertFull();
        } else {
            b = advertIncr();
        }
        if (b) {
            conn.needFull++;
            return;
        }
        if (conn.needFull != 0) {
            doing--;
        }
        conn.adversion = doing;
    }

    /**
     * set accepted list
     */
    public void setAccepted() {
        accUni = new tabRoute<addrIP>("bgp");
        accMlt = new tabRoute<addrIP>("bgp");
        accFlw = new tabRoute<addrIP>("bgp");
        accVpnU = new tabRoute<addrIP>("bgp");
        accVpnM = new tabRoute<addrIP>("bgp");
        accVpnF = new tabRoute<addrIP>("bgp");
        accVpls = new tabRoute<addrIP>("bgp");
        accEvpn = new tabRoute<addrIP>("bgp");
        accMdt = new tabRoute<addrIP>("bgp");
        reachable = false;
        if (lower.nhtRoumap != null) {
            tabRouteEntry<addrIP> rou = lower.fwdCore.actualR.route(peerAddr);
            if (rou == null) {
                return;
            }
            tabRtrmapN rme = lower.nhtRoumap.find(rou);
            if (rme == null) {
                return;
            }
            if (rme.action != tabPlcmapN.actionType.actPermit) {
                return;
            }
        }
        if (lower.nhtRouplc != null) {
            tabRouteEntry<addrIP> rou = lower.fwdCore.actualR.route(peerAddr);
            if (rou == null) {
                return;
            }
            rou = tabRtrplc.doRpl(rou, lower.nhtRouplc, true);
            if (rou == null) {
                return;
            }
        }
        if (lower.nhtPfxlst != null) {
            tabRouteEntry<addrIP> rou = lower.fwdCore.actualR.route(peerAddr);
            if (rou == null) {
                return;
            }
            tabPrfxlstN ple = lower.nhtPfxlst.find(rou);
            if (ple == null) {
                return;
            }
            if (ple.action != tabPlcmapN.actionType.actPermit) {
                return;
            }
        }
        reachable = true;
        if (!softReconfig) {
            accUni = conn.lrnUni;
            accMlt = conn.lrnMlt;
            accFlw = conn.lrnFlw;
            accVpnU = conn.lrnVpnU;
            accVpnM = conn.lrnVpnM;
            accVpnF = conn.lrnVpnF;
            accVpls = conn.lrnVpls;
            accEvpn = conn.lrnEvpn;
            accMdt = conn.lrnMdt;
            return;
        }
        tabRoute.addUpdatedTable(2, accUni, conn.lrnUni, roumapIn, roupolIn, prflstIn);
        tabRoute.addUpdatedTable(2, accMlt, conn.lrnMlt, roumapIn, roupolIn, prflstIn);
        tabRoute.addUpdatedTable(2, accFlw, conn.lrnFlw, voumapIn, voupolIn, null);
        tabRoute.addUpdatedTable(2, accVpnU, conn.lrnVpnU, voumapIn, voupolIn, null);
        tabRoute.addUpdatedTable(2, accVpnM, conn.lrnVpnM, voumapIn, voupolIn, null);
        tabRoute.addUpdatedTable(2, accVpnF, conn.lrnVpnF, voumapIn, voupolIn, null);
        tabRoute.addUpdatedTable(2, accVpls, conn.lrnVpls, voumapIn, voupolIn, null);
        tabRoute.addUpdatedTable(2, accEvpn, conn.lrnEvpn, voumapIn, voupolIn, null);
        tabRoute.addUpdatedTable(2, accMdt, conn.lrnMdt, voumapIn, voupolIn, null);
    }

    /**
     * set merged list
     *
     * @param uni unicast
     * @param mlt multicast
     * @param flw flowspec
     * @param vpnU vpn uni
     * @param vpnM vpn multi
     * @param vpnF vpn flow
     * @param vpls vpls
     * @param evpn evpn
     * @param mdt mdt
     */
    public void setMerge(tabRoute<addrIP> uni, tabRoute<addrIP> mlt, tabRoute<addrIP> flw, tabRoute<addrIP> vpnU, tabRoute<addrIP> vpnM, tabRoute<addrIP> vpnF, tabRoute<addrIP> vpls, tabRoute<addrIP> evpn, tabRoute<addrIP> mdt) {
        uni.mergeFrom(2, new tabRoute<addrIP>(accUni), null, false);
        mlt.mergeFrom(2, new tabRoute<addrIP>(accMlt), null, false);
        flw.mergeFrom(2, new tabRoute<addrIP>(accFlw), null, false);
        vpnU.mergeFrom(2, new tabRoute<addrIP>(accVpnU), null, false);
        vpnM.mergeFrom(2, new tabRoute<addrIP>(accVpnM), null, false);
        vpnF.mergeFrom(2, new tabRoute<addrIP>(accVpnF), null, false);
        vpls.mergeFrom(2, new tabRoute<addrIP>(accVpls), null, false);
        evpn.mergeFrom(2, new tabRoute<addrIP>(accEvpn), null, false);
        mdt.mergeFrom(2, new tabRoute<addrIP>(accMdt), null, false);
    }

    /**
     * set group membership
     */
    public void setGroup() {
        if (reachable != reachOld) {
            reachOld = reachable;
            reachTim = bits.getTime();
            reachNum++;
            if (debugger.rtrBgpEvnt) {
                logger.debug("reachable neighbor " + peerAddr + " " + reachable);
            }
        }
        groupMember = -1;
        if (shutdown) {
            return;
        }
        for (int i = 0; i < lower.groups.size(); i++) {
            rtrBgpGroup ntry = lower.groups.get(i);
            if (ntry.peerType != peerType) {
                continue;
            }
            if (localAddr.compare(localAddr, ntry.localAddr) != 0) {
                continue;
            }
            if (ntry.sameOutput(this)) {
                continue;
            }
            groupMember = i;
            return;
        }
        groupMember = lower.groups.size();
        rtrBgpGroup ntry = new rtrBgpGroup(lower, groupMember);
        ntry.copyFrom(this);
        ntry.peerType = peerType;
        ntry.localAddr = localAddr.copyBytes();
        lower.groups.add(ntry);
    }

    /**
     * set needed prefixes
     */
    public void setNeeded() {
        if (groupMember < 0) {
            wilUni = new tabRoute<addrIP>("tx");
            wilMlt = new tabRoute<addrIP>("tx");
            wilFlw = new tabRoute<addrIP>("tx");
            wilVpnU = new tabRoute<addrIP>("tx");
            wilVpnM = new tabRoute<addrIP>("tx");
            wilVpnF = new tabRoute<addrIP>("tx");
            wilVpls = new tabRoute<addrIP>("tx");
            wilEvpn = new tabRoute<addrIP>("tx");
            wilMdt = new tabRoute<addrIP>("tx");
            chgUni = new tabRoute<addrIP>("chg");
            chgMlt = new tabRoute<addrIP>("chg");
            chgFlw = new tabRoute<addrIP>("chg");
            chgVpnU = new tabRoute<addrIP>("chg");
            chgVpnM = new tabRoute<addrIP>("chg");
            chgVpnF = new tabRoute<addrIP>("chg");
            chgVpls = new tabRoute<addrIP>("chg");
            chgEvpn = new tabRoute<addrIP>("chg");
            chgMdt = new tabRoute<addrIP>("chg");
        } else {
            rtrBgpGroup grp = lower.groups.get(groupMember);
            wilUni = grp.wilUni;
            wilMlt = grp.wilMlt;
            wilFlw = grp.wilFlw;
            wilVpnU = grp.wilVpnU;
            wilVpnM = grp.wilVpnM;
            wilVpnF = grp.wilVpnF;
            wilVpls = grp.wilVpls;
            wilEvpn = grp.wilEvpn;
            wilMdt = grp.wilMdt;
            chgUni = grp.chgUni;
            chgMlt = grp.chgMlt;
            chgFlw = grp.chgFlw;
            chgVpnU = grp.chgVpnU;
            chgVpnM = grp.chgVpnM;
            chgVpnF = grp.chgVpnF;
            chgVpls = grp.chgVpls;
            chgEvpn = grp.chgEvpn;
            chgMdt = grp.chgMdt;
        }
        conn.needFull++;
    }

    private void setValidityTable(tabRoute<addrIP> tab) {
        for (int i = 0; i < tab.size(); i++) {
            lower.setValidity(tab.get(i));
        }
    }

    /**
     * validate prefixes
     */
    public void setValidity() {
        if (lower.rpkis.size() < 1) {
            return;
        }
        setValidityTable(accUni);
        setValidityTable(accMlt);
    }

    /**
     * group version
     */
    public void setGrpVer() {
        if (groupMember < 0) {
            return;
        }
        if (conn.initFull) {
            return;
        }
        if (conn.needFull > 1) {
            return;
        }
        rtrBgpGroup grp = lower.groups.get(groupMember);
        int i = conn.adversion;
        if (i < grp.minversion) {
            grp.minversion = i;
        }
    }

    /**
     * get accepted
     *
     * @param safi safi to query
     * @return table
     */
    public tabRoute<addrIP> getAccepted(int safi) {
        if (safi == lower.afiUni) {
            return accUni;
        }
        if (safi == lower.afiLab) {
            return accUni;
        }
        if (safi == lower.afiMlt) {
            return accMlt;
        }
        if (safi == lower.afiFlw) {
            return accFlw;
        }
        if (safi == lower.afiVpnU) {
            return accVpnU;
        }
        if (safi == lower.afiVpnM) {
            return accVpnM;
        }
        if (safi == lower.afiVpnF) {
            return accVpnF;
        }
        if (safi == lower.afiVpls) {
            return accVpls;
        }
        if (safi == lower.afiEvpn) {
            return accEvpn;
        }
        if (safi == lower.afiMdt) {
            return accMdt;
        }
        return null;
    }

    /**
     * get willing
     *
     * @param safi safi to query
     * @return table
     */
    public tabRoute<addrIP> getWilling(int safi) {
        if (safi == lower.afiUni) {
            return wilUni;
        }
        if (safi == lower.afiLab) {
            return wilUni;
        }
        if (safi == lower.afiMlt) {
            return wilMlt;
        }
        if (safi == lower.afiFlw) {
            return wilFlw;
        }
        if (safi == lower.afiVpnU) {
            return wilVpnU;
        }
        if (safi == lower.afiVpnM) {
            return wilVpnM;
        }
        if (safi == lower.afiVpnF) {
            return wilVpnF;
        }
        if (safi == lower.afiVpls) {
            return wilVpls;
        }
        if (safi == lower.afiEvpn) {
            return wilEvpn;
        }
        if (safi == lower.afiMdt) {
            return wilMdt;
        }
        return null;
    }

    /**
     * neighbor list entry
     *
     * @param safi safi to query
     * @return line of string
     */
    public String showNeighs(int safi) {
        return remoteAs + "|" + tabSiz(conn.getLearned(safi)) + "|" + tabSiz(getAccepted(safi)) + "|" + tabSiz(getWilling(safi)) + "|" + tabSiz(conn.getAdverted(safi)) + "|" + peerAddr + "|" + bits.timePast(conn.upTime);
    }

    /**
     * neighbor list entry
     *
     * @return line of string
     */
    public String showNeighs() {
        return remoteAs + "|" + rtrBgpParam.mask2string(conn.peerAfis) + "|" + peerAddr + "|" + bits.timePast(conn.upTime);
    }

    private String tabSiz(tabRoute<addrIP> tab) {
        if (tab == null) {
            return "-";
        }
        return "" + tab.size();
    }

    /**
     * group list entry
     *
     * @return line of string
     */
    public String showGroups() {
        return remoteAs + "|" + groupMember + "|" + peerAddr + "|" + bits.timePast(conn.upTime);
    }

    /**
     * nexthop list entry
     *
     * @return line of string
     */
    public String showNxtHps() {
        return remoteAs + "|" + reachable + "|" + bits.timePast(reachTim) + "|" + reachNum + "|" + peerAddr + "|" + bits.timePast(conn.upTime);
    }

    /**
     * graceful restart list entry
     *
     * @return line of string
     */
    public String showGrcRst() {
        return remoteAs + "|" + rtrBgpParam.mask2string(conn.peerGrace) + "|" + rtrBgpParam.mask2string(graceRestart & addrFams) + "|" + peerAddr;
    }

    /**
     * additional path list entry
     *
     * @return line of string
     */
    public String showAddPth() {
        return remoteAs + "|" + rtrBgpParam.mask2string(conn.addpathRx) + "|" + rtrBgpParam.mask2string(conn.addpathTx) + "|" + peerAddr;
    }

    /**
     * router id list entry
     *
     * @return line of string
     */
    public String showRtrId() {
        return remoteAs + "|" + conn.peerRouterID + "|" + conn.peer32bitAS + "|" + conn.peerRefresh + "|" + peerAddr;
    }

    /**
     * buffer list entry
     *
     * @return line of string
     */
    public String showBuffer() {
        return remoteAs + "|" + pipeSide.getStatus(conn.pipe) + "|" + conn.cntr.getShStat() + "|" + peerAddr;
    }

}
